@@ -14,21 +14,22 @@
 
 #define sexp_import_prefix "(import ("
 #define sexp_import_suffix "))"
-#define sexp_environment_prefix "(environment '("
+#define sexp_environment_prefix "(mutable-environment '("
 #define sexp_environment_suffix "))"
 #define sexp_trace_prefix "(module-env (load-module '("
 #define sexp_trace_suffix ")))"
-#define sexp_default_environment "(environment '(scheme small))"
+#define sexp_default_environment "(mutable-environment '(scheme small))"
 #define sexp_advice_environment "(load-module '(chibi repl))"
 
 #define sexp_version_string "chibi-scheme "sexp_version" \""sexp_release_name"\" "
 
 #ifdef PLAN9
 #define exit_failure() exits("ERROR")
+#define exit_success() exits(nil)
 #else
 #define exit_failure() exit(70)
-#endif
 #define exit_success() exit(0)
+#endif
 
 #if SEXP_USE_MAIN_HELP
 void sexp_usage(int err) {
@@ -186,14 +187,12 @@ static sexp_uint_t multiplier (char c) {
 #endif
 
 static char* make_import(const char* prefix, const char* mod, const char* suffix) {
-  int preflen = strlen(prefix), modlen = strlen(mod);
-  int len = preflen + modlen + strlen(suffix);
-  int suflen = strlen(suffix) + (mod[0] == '(' ? 1 : 0);
-  char *p, *impmod = (char*) malloc(len+1);
-  snprintf(impmod, len, "%s", prefix);
-  snprintf(impmod+preflen, len-preflen, "%s", mod[0] == '(' ? mod + 1 : mod);
-  snprintf(impmod+len-suflen, suflen+1, "%s", suffix);
-  impmod[len] = '\0';
+  char *impmod, *p;
+  size_t impmodsize;
+  if (mod[0] == '(') mod++;
+  impmodsize = strlen(prefix) + strlen(mod) + strlen(suffix) + 1;
+  impmod = (char*) malloc(impmodsize);
+  snprintf(impmod, impmodsize, "%s%s%s", prefix, mod, suffix);
   for (p=impmod; *p; p++)
     if (*p == '.') *p=' ';
   return impmod;
@@ -243,7 +242,7 @@ static sexp sexp_add_import_binding (sexp ctx, sexp env) {
   tmp = sexp_env_ref(ctx, sexp_meta_env(ctx), sym, SEXP_VOID);
   sym = sexp_intern(ctx, "import", -1);
   sexp_env_define(ctx, env, sym, tmp);
-  sexp_gc_release3(ctx);
+  sexp_gc_release2(ctx);
   return env;
 }
 
@@ -290,7 +289,7 @@ static void do_init_context (sexp* ctx, sexp* env, sexp_uint_t heap_size,
 
 #define load_init(bootp) if (! init_loaded++) do {                      \
       init_context();                                                   \
-      check_exception(ctx, env=sexp_load_standard_repl_env(ctx, env, SEXP_SEVEN, bootp, nonblocking)); \
+      check_exception(ctx, env=sexp_load_standard_repl_env(ctx, env, standard, bootp, nonblocking)); \
     } while (0)
 
 /* static globals for the sake of resuming from within emscripten */
@@ -308,7 +307,7 @@ sexp run_main (int argc, char **argv) {
   sexp_sint_t i, j, c, quit=0, print=0, init_loaded=0, mods_loaded=0,
     fold_case=SEXP_DEFAULT_FOLD_CASE_SYMS, nonblocking=0;
   sexp_uint_t heap_size=0, heap_max_size=SEXP_MAXIMUM_HEAP_SIZE;
-  sexp out=SEXP_FALSE, ctx=NULL, ls;
+  sexp out=SEXP_FALSE, ctx=NULL, ls, res=SEXP_ZERO, standard=SEXP_SEVEN;
   sexp_gc_var4(tmp, sym, args, env);
   args = SEXP_NULL;
   env = NULL;
@@ -505,6 +504,15 @@ sexp run_main (int argc, char **argv) {
     case 'r':
       main_symbol = argv[i][2] == '\0' ? "main" : argv[i]+2;
       break;
+    case 'S':
+      arg = ((argv[i][2] == '\0') ? argv[++i] : argv[i]+2);
+      j = atoi(arg);
+      if (0 < j && j <= 1000) {
+        standard = sexp_make_fixnum(j);
+      } else {
+        fprintf(stderr, "-S<standard> should be an integer in [1, 1000] but got %s\n", arg);
+      }
+      break;
     case 's':
       init_context(); sexp_global(ctx, SEXP_G_STRICT_P) = SEXP_TRUE;
       handle_noarg();
@@ -562,7 +570,7 @@ sexp run_main (int argc, char **argv) {
       sym=sexp_intern(ctx, sexp_raw_script_file_symbol, -1), tmp);
     for (j=argc-1; j>=i; j--)
       args = sexp_cons(ctx, tmp=sexp_c_string(ctx,argv[j],-1), args);
-    if (main_symbol)
+    if (main_module)
       args = sexp_cons(ctx, tmp=sexp_c_string(ctx,main_symbol,-1), args);
     if (args == SEXP_NULL)
       args = sexp_cons(ctx, tmp=sexp_c_string(ctx,"",-1), args);
@@ -639,7 +647,7 @@ sexp run_main (int argc, char **argv) {
         tmp = sexp_env_ref(ctx, env, sym, SEXP_FALSE);
         if (sexp_procedurep(tmp)) {
           args = sexp_list1(ctx, args);
-          check_exception(ctx, sexp_apply(ctx, tmp, args));
+          res = check_exception(ctx, sexp_apply(ctx, tmp, args));
         } else {
           fprintf(stderr, "couldn't find main binding: %s in %s\n", main_symbol, main_module ? main_module : argv[i]);
         }
@@ -652,7 +660,7 @@ sexp run_main (int argc, char **argv) {
     fprintf(stderr, "destroy_context error\n");
     return SEXP_FALSE;
   }
-  return SEXP_TRUE;
+  return res;
 }
 
 #ifdef EMSCRIPTEN
@@ -668,11 +676,24 @@ void sexp_resume() {
 #endif
 
 int main (int argc, char **argv) {
+  sexp res;
 #if SEXP_USE_PRINT_BACKTRACE_ON_SEGFAULT
   signal(SIGSEGV, sexp_segfault_handler); 
 #endif
   sexp_scheme_init();
-  if (run_main(argc, argv) == SEXP_FALSE) {
+  res = run_main(argc, argv);
+  if (sexp_fixnump(res)) {
+    int code = sexp_unbox_fixnum(res);
+#ifdef PLAN9
+    if (code == 0) {
+      exit_success();
+    } else {
+      exit_failure();
+    }
+#else
+    return code;
+#endif
+  } else if (res == SEXP_FALSE) {
     exit_failure();
   } else {
     exit_success();
