@@ -787,18 +787,19 @@ static sexp make_param_list (sexp ctx, sexp_uint_t i) {
   return res;
 }
 
-static sexp make_opcode_procedure (sexp ctx, sexp op, sexp_uint_t i) {
+static sexp make_opcode_procedure (sexp ctx, sexp op, sexp_uint_t i, sexp_sint_t flags) {
+  int j = i+(flags & SEXP_PROC_VARIADIC);
   sexp ls, res, env;
   sexp_gc_var6(bc, params, ref, refs, lambda, ctx2);
-  if (i == sexp_opcode_num_args(op)) { /* return before preserving */
+  if (j == sexp_opcode_num_args(op)) { /* return before preserving */
     if (sexp_opcode_proc(op)) return sexp_opcode_proc(op);
-  } else if (i < sexp_opcode_num_args(op)) {
+  } else if (j < sexp_opcode_num_args(op)) {
     return sexp_compile_error(ctx, "not enough args for opcode", op);
   } else if (! sexp_opcode_variadic_p(op)) { /* i > num_args */
     return sexp_compile_error(ctx, "too many args for opcode", op);
   }
   sexp_gc_preserve6(ctx, bc, params, ref, refs, lambda, ctx2);
-  params = make_param_list(ctx, i);
+  params = make_param_list(ctx, j);
   lambda = sexp_make_lambda(ctx, params);
   ctx2 = sexp_make_child_context(ctx, lambda);
   env = sexp_extend_env(ctx2, sexp_context_env(ctx), params, lambda);
@@ -819,8 +820,8 @@ static sexp make_opcode_procedure (sexp ctx, sexp op, sexp_uint_t i) {
       generate_opcode_app(ctx2, refs);
       bc = sexp_complete_bytecode(ctx2);
       sexp_bytecode_name(bc) = sexp_opcode_name(op);
-      res=sexp_make_procedure(ctx2, SEXP_ZERO, sexp_make_fixnum(i), bc, SEXP_VOID);
-      if (i == sexp_opcode_num_args(op))
+      res=sexp_make_procedure(ctx2, sexp_make_fixnum(flags), sexp_make_fixnum(i), bc, SEXP_VOID);
+      if (j == sexp_opcode_num_args(op))
         sexp_opcode_proc(op) = res;
     }
   }
@@ -828,6 +829,28 @@ static sexp make_opcode_procedure (sexp ctx, sexp op, sexp_uint_t i) {
   return res;
 }
 
+sexp sexp_make_foreign_proc(sexp ctx, const char *name, int num_args, int flags,
+                            const char *fname, sexp_proc1 f) {
+  sexp_gc_var1(res);
+  sexp_gc_preserve1(ctx, res);
+  res = sexp_make_foreign (ctx, name, num_args+((flags & SEXP_PROC_VARIADIC)>0), 0, fname, f, NULL);
+  if (!sexp_exceptionp(res))
+    res = make_opcode_procedure (ctx, res, num_args, flags);
+  sexp_gc_release1(ctx);
+  return res;
+}
+
+sexp sexp_define_foreign_proc_aux (sexp ctx, sexp env, const char *name,int num_args,
+                                   int flags, const char *fname, sexp_proc1 f, sexp data) {
+  sexp_gc_var2(sym, res);
+  sexp_gc_preserve2(ctx, sym, res);
+  res = sexp_make_foreign_proc(ctx, name, num_args, flags, fname, f);
+  if (!sexp_exceptionp(res))
+    sexp_env_define(ctx, env, sym = sexp_intern(ctx, name, -1), res);
+  sexp_gc_release2(ctx);
+  return res;
+}
+
 /*********************** the virtual machine **************************/
 
 sexp sexp_make_trampoline (sexp ctx, sexp proc, sexp args) {
@@ -1032,6 +1055,7 @@ static void* sexp_thread_debug_event(sexp ctx) {
 #if SEXP_USE_POLL_PORT
 int sexp_poll_port(sexp ctx, sexp port, int inputp) {
   fd_set fds;
+  struct timeval timeout;
   int fd = sexp_port_fileno(port);
   if (fd < 0) {
     usleep(SEXP_POLL_SLEEP_TIME);
@@ -1039,7 +1063,9 @@ int sexp_poll_port(sexp ctx, sexp port, int inputp) {
   }
   FD_ZERO(&fds);
   FD_SET(fd, &fds);
-  return select(1, (inputp ? &fds : NULL), (inputp ? NULL : &fds), NULL, NULL);
+  timeout.tv_sec = 0;
+  timeout.tv_usec = 10000;  /* 10millis */
+  return select(1, (inputp ? &fds : NULL), (inputp ? NULL : &fds), NULL, &timeout);
 }
 #endif
 
@@ -1069,7 +1095,9 @@ sexp sexp_apply (sexp ctx, sexp proc, sexp args) {
   sexp_ensure_stack(i + 64 + (sexp_procedurep(tmp1) ? sexp_bytecode_max_depth(sexp_procedure_code(tmp1)) : 0));
   for (top += i; sexp_pairp(tmp2); tmp2=sexp_cdr(tmp2), top--)
     _ARG1 = sexp_car(tmp2);
-  top += i+1;
+  top += i;
+  /* restore the make_call invariant */
+  _PUSH(tmp1);
   goto make_call;
 
  loop:
@@ -1240,7 +1268,9 @@ sexp sexp_apply (sexp ctx, sexp proc, sexp args) {
       int prev_top = top;
       for (top=fp-j+i-1; sexp_pairp(tmp2); tmp2=sexp_cdr(tmp2), top--)
         stack[top] = sexp_car(tmp2);
-      top = fp+i-j+1;
+      top = fp+i-j;
+      /* restore the make_call invariant */
+      _PUSH(tmp1);
       fp = k;
       /* if final cdr of tmp2 isn't null, then args list was improper */
       if (! sexp_nullp(tmp2)) {
@@ -1263,7 +1293,9 @@ sexp sexp_apply (sexp ctx, sexp proc, sexp args) {
     /* copy new args into place */
     for (k=0; k<i; k++)
       stack[fp-j+k] = stack[top-1-i+k];
-    top = fp+i-j+1;
+    top = fp+i-j;
+    /* restore the make_call invariant */
+    _PUSH(tmp1);
     fp = sexp_unbox_fixnum(tmp2);
     goto make_call;
   case SEXP_OP_CALL:
@@ -1274,7 +1306,7 @@ sexp sexp_apply (sexp ctx, sexp proc, sexp args) {
     sexp_context_top(ctx) = top;
     if (sexp_opcodep(tmp1)) {
       /* compile non-inlined opcode applications on the fly */
-      tmp1 = make_opcode_procedure(ctx, tmp1, i);
+      tmp1 = make_opcode_procedure(ctx, tmp1, i, SEXP_PROC_NONE);
       if (sexp_exceptionp(tmp1)) {
         _ARG1 = tmp1;
         goto call_error_handler;
@@ -1865,6 +1897,9 @@ sexp sexp_apply (sexp ctx, sexp proc, sexp args) {
       if (tmp2 == SEXP_ZERO)
         sexp_raise("divide by zero", SEXP_NULL);
       _ARG1 = sexp_fx_div(tmp1, tmp2);
+      if ((sexp_sint_t)tmp1 < 0 && (sexp_sint_t)tmp2 < 0 && (sexp_sint_t)_ARG1 < 0) {
+        _ARG1 = sexp_quotient(ctx, tmp1=sexp_fixnum_to_bignum(ctx, tmp1), tmp2);
+      }
     }
 #if SEXP_USE_BIGNUMS
     else {
@@ -2065,7 +2100,7 @@ sexp sexp_apply (sexp ctx, sexp proc, sexp args) {
     else
 #endif
     i = sexp_write_char(ctx, sexp_unbox_character(_ARG1), _ARG2);
-    if (i == EOF) {
+    if ((int)i == EOF) {
       if (!sexp_port_openp(_ARG2))
         sexp_raise("write-char: port is closed", _ARG2);
       else
@@ -2144,7 +2179,7 @@ sexp sexp_apply (sexp ctx, sexp proc, sexp args) {
     errno = 0;
 #endif
     i = sexp_read_char(ctx, _ARG1);
-    if (i == EOF) {
+    if ((int)i == EOF) {
       if (!sexp_port_openp(_ARG1)) {
         sexp_raise("read-char: port is closed", _ARG1);
 #if SEXP_USE_GREEN_THREADS
@@ -2180,7 +2215,7 @@ sexp sexp_apply (sexp ctx, sexp proc, sexp args) {
     errno = 0;
 #endif
     i = sexp_read_char(ctx, _ARG1);
-    if (i == EOF) {
+    if ((int)i == EOF) {
       if (!sexp_port_openp(_ARG1))
         sexp_raise("peek-char: port is closed", _ARG1);
       else
